# -*- coding: UTF-8 -*-
# pylint:disable=missing-docstring,import-error,bad-whitespace

"""API to access the BPE over RPC"""

# System imports
from django              import db
from django.forms.models import model_to_dict

# Project imports
from symonitoring_bpe       import models
from symonitoring_bpe       import common as cmn
from symonitoring_model     import models as models_sm
from symonitoring_rpcserver import livestatus as ls

from symonitoring_rpcserver.apiregistry import api_callable
from functools import wraps


# a decorator to force db reconnection on error
def reconnect_db_on_error(method):
    """Reconnects to the database when a DB
    error occurs"""
    @wraps(method)
    def wrapper(*args, **kwargs):
        try:
            return method(*args, **kwargs)
        except db.utils.DatabaseError as exc:
            cmn.lg.exception((
                'API encountered a database exception: %s. '
                'Reconnecting.'
            ) % exc)
            db.connection.close()
            db.close_old_connections()
            return method(*args, **kwargs)
    return wrapper


# "private" methods functions
def _get_hosts(component):
    """Return a list of hosts assigned to the component

    The list items are dicts containing essential information.
    """
    # TODO: do some logging?
    return [{
        'id':           ph.host.pk,
        'parent':       ph.component.pk,
        'name':         ph.host.hostname,
        'state':        ph.state,
        'type':         'host',
    } for ph in component.processhost_set.all()]


def _get_services(component):
    """Return a list of services assigned to the component

    The list items are dicts containing essential information.
    """
    # TODO: do some logging?
    return [{
        'id':           ps.service.pk,
        'parent':       ps.component.pk,
        'name':         ps.service.display_name,
        'host_id':      ps.service.hostserviceid.hostid.pk,
        'host_name':    ps.service.hostserviceid.hostid.hostname,
        'state':        ps.state,
        'type':         'service'
    } for ps in component.processservice_set.all()]


def _get_applications(component):
    """Return a list of applications assigned to the component

    The list items are dicts containing essential information.
    """
    # TODO: do some logging?
    return [{
        'id':             a.pk,
        'customer_name':  a.customer.name,
        'parent':         component.pk,
        'name':           a.name,
        'description':    a.description,
        'state':          a.state,
        'type':           'application',
    } for a in component.applications.all()]


def _get_components(component_or_app, parent=None):
    """Return component tree as a list of dicts

    Recursion, see Recursion.

    Please bear in mind that this algorithm calculates the
    availability of each component by following the path from
    component to component. It will only consider assigned hosts
    and services attached to the last component in the path. It
    doesn't make sense to have both, child components and hosts or
    services assigned.
    """
    # TODO: do some logging?
    return [{
        'id':               c.pk,
        'application':      c.application.pk,
        'parent':           c.parent and c.parent.pk,
        'name':             c.name,
        'description':      c.description,
        'min_available':    c.min_available,
        'hosts':            _get_hosts(c),
        'services':         _get_services(c),
        'applications':     _get_applications(c),
        'components':       _get_components(c, c.pk),
        'state':            c.state,
        'type':             'component'
    } for c in component_or_app.processcomponent_set.filter(
        parent=parent
    )]


# api callable functions
@api_callable
@reconnect_db_on_error
def get_application_list(customer_ids=None, application_id=None):
    """Returns a list of available applications"""
    cmn.lg.debug('Getting application list for customers: %s' % customer_ids)
    applications = models.ProcessApplication.objects.all()
    # if the function argumnents are given, do some further filtering
    if customer_ids:
        applications = applications.filter(customer__in=customer_ids)
    if application_id:
        # if an application id is given, exclude the applications which can't
        # be added to the respective application
        application = models.ProcessApplication.objects.get(
            pk=application_id
        )
        applications = applications.exclude(
            pk__in=application.subapplications()
        ).exclude(
            pk__in=application.superapplications()
        )
    # sort the applications by name
    applications.order_by('name')
    cmn.lg.debug('Got %s applications', len(applications))
    return [model_to_dict(a) for a in applications]


@api_callable
@reconnect_db_on_error
def get_component_list(customer_ids=None):
    """Returns a list of available applications"""
    cmn.lg.debug('Getting component list for customers: %s' % customer_ids)
    components = models.ProcessComponent.objects.all()
    # if some customer's ids are given, pick only components for the customer
    if customer_ids:
        components = components.filter(
            application__customer__id__in=customer_ids
        )
    # sort the components by name
    components.order_by('name')
    cmn.lg.debug('Got %s components', len(components))
    return [model_to_dict(c) for c in components]


@api_callable
@reconnect_db_on_error
def get_application(application_id):
    """Returns the components of an application."""
    cmn.lg.debug('Getting application %s' % application_id)
    application = models.ProcessApplication.objects.get(
        pk=application_id
    )
    return {
        'id':               application.pk,
        'name':             application.name,
        'customer':         application.customer.pk,
        'description':      application.description,
        'min_available':    application.min_available,
        'components':       _get_components(application),
        'state':            application.state,
        'type':             'root'
    }


@api_callable
def create_application(name, customer_id, min_available, description):
    """Create new applicaton"""
    application = models.ProcessApplication()
    application.name = name
    application.min_available = min_available
    application.customer = models_sm.Customer.objects.get(
        pk=customer_id
    )
    application.description = description
    application.save()
    return model_to_dict(application)


@api_callable
def update_application(
    application_id,
    name,
    customer_id,
    min_available,
    description
):
    """Update an application"""
    application = models.ProcessApplication.objects.get(pk=application_id)
    application.name = name
    application.min_available = min_available
    application.customer = models_sm.Customer.objects.get(
        pk=customer_id
    )
    application.description = description
    application.save()
    return model_to_dict(application)


@api_callable
def delete_application(application_id):
    """Delete an application"""
    models.ProcessApplication.objects.get(
        pk=application_id
    ).delete()


@api_callable
def create_component(
    name,
    min_available,
    application_id,
    parent_id,
    description
):
    """Create new component"""
    component = models.ProcessComponent()
    component.application = models.ProcessApplication.objects.get(
        pk=application_id
    )
    if parent_id:
        component.parent = models.ProcessComponent.objects.get(
            pk=parent_id
        )
    component.name = name
    component.min_available = min_available
    component.description = description
    component.save()
    return model_to_dict(component)


@api_callable
def update_component(
    component_id,
    name,
    min_available,
    application_id,
    parent_id,
    description
):
    """Update component by given primary key."""
    # get the component
    component = models.ProcessComponent.objects.get(
        pk=component_id
    )
    # edit the component
    component.name          = name
    component.min_available = min_available
    component.application   = models.ProcessApplication.objects.get(
        pk=application_id
    )
    component.parent        = models.ProcessComponent.objects.filter(
        pk=parent_id
    ).first()
    component.description   = description
    # save the component
    component.save()
    return model_to_dict(component)


@api_callable
def get_component(component_id):
    """Returns a specific component by ID.

    :param component_id: the identifier of the component
    :type  component_id: int

    :returns: component
    :rtype:   models.ProcessComponent
    """

    cmn.lg.debug('Getting component %d' % component_id)

    try:
        component = models.ProcessComponent.objects.get(  # noqa pylint: disable=no-member
            pk=component_id
        )
    except models.ProcessComponent.DoesNotExist:  # noqa pylint: disable=no-member
        raise IndexError('Component with ID {id} does not exist'.format(
            id=component_id
        ))

    return model_to_dict(component)


@api_callable
def delete_component(component_id):
    """Delete component"""
    models.ProcessComponent.objects.get(
        pk=component_id
    ).delete()
    # TODO: - remove all the hosts and services linked to this component


@api_callable
def add_service_to_component(component_id, service_id):
    """Add a service to a component

    The parameter instance_id is the service id.
    """
    cmn.lg.debug(
        'Got request for component %d service %d',
        component_id,
        service_id
    )
    component = models.ProcessComponent.objects.get(
        pk=component_id
    )
    instance = models_sm.Hostserviceinstance.objects.get(
        pk=service_id
    )
    process_service = models.ProcessService()
    process_service.component = component
    process_service.service = instance
    process_service.save()
    return model_to_dict(process_service)


@api_callable
def add_host_to_component(component_id, host_id):
    """Add a host to a component"""
    host = models_sm.Host.objects.get(
        pk=host_id
    )
    component = models.ProcessComponent.objects.get(
        pk=component_id
    )
    process_host = models.ProcessHost()
    process_host.component = component
    process_host.host = host
    process_host.save()
    return model_to_dict(process_host)


@api_callable
def add_application_to_component(component_id, application_id):
    """Add an application to a component

    The parameter application_id is the application id.
    """
    cmn.lg.debug(
        'Got request for component %d application %d',
        component_id,
        application_id
    )
    component = models.ProcessComponent.objects.get(
        pk=component_id
    )
    application = models.ProcessApplication.objects.get(
        pk=application_id
    )
    component.applications.add(application)


@api_callable
def remove_service_from_component(service_id, component_id):
    """Remove a service from a component"""
    service = models.ProcessService.objects.filter(
        component_id=component_id,
        service_id=service_id
    ).first()
    service.delete()


@api_callable
def remove_host_from_component(host_id, component_id):
    """Remove a host from a component"""
    host = models.ProcessHost.objects.filter(
        component_id=component_id,
        host_id=host_id
    ).first()
    host.delete()


@api_callable
def remove_application_from_component(application_id, component_id):
    """Remove a host from a component"""
    app = models.ProcessApplication.objects.get(id=application_id)
    component = models.ProcessComponent.objects.get(id=component_id)
    component.applications.remove(app)


@api_callable
@reconnect_db_on_error
def get_host_list_by_ldap_group(ldap_group):
    """Returns a list of hosts belonging to an LDAP group

    LDAP groups can be mapped to customers in SyMonitoring. This
    method looks up the customers linked to a given LDAP group and
    returns all hosts belonging to these customers.
    """
    cmn.lg.debug('Getting host list for LDAP group %s' % ldap_group)
    return [{
        'id': h.pk,
        'name': h.hostname,
        'description': h.description,
        'ipv4': h.ipv4,
        'ipv6': h.ipv6,
        'type': h.typeid and {
            'id': h.typeid.pk,
            'name': h.typeid.name,
        } or None,
        'customer': {
            'id':   h.customerid.pk,
            'name': h.customerid.name,
            'sla':  {
                'id':   h.customerid.slaid.pk,
                'name': h.customerid.slaid.name,
            },
        },
        'os': {
            'id':   h.osid.pk,
            'name': h.osid.name,
        },
        'os_version': h.osversionid and {
            'id':     h.osversionid.pk,
            'name':   h.osversionid.name,
        } or None,
        'isenabled': h.isenabled,
        'state':     h.isenabled and ls.get_host_state(h) or None,
    } for h in models_sm.Host.objects.filter(
        isenabled=True,
        customerid__ldapgroup=ldap_group
    )]


@api_callable
@reconnect_db_on_error
def get_service_list_by_host(host_id):
    """Returns a list of services assigned to given host"""
    cmn.lg.debug('Getting service list for host %s' % host_id)
    # TODO: add this to the dict
    return [{
        'id':        hsi.pk,
        'name':      hsi.display_name,
        'isenabled': hsi.isenabled,
        'host':      {
            'id':   hsi.hostserviceid.hostid.pk,
            'name': hsi.hostserviceid.hostid.hostname,
        },
        'sla':       hsi.slaid and {
            'id':   hsi.slaid.pk,
            'name': hsi.slaid.name,
        } or None,
        'state':     hsi.isenabled and ls.get_service_state(hsi) or None,
    } for hsi in models_sm.Hostserviceinstance.objects.filter(
        isenabled=True,
        hostserviceid__hostid=host_id,
        hostserviceid__hostid__isenabled=True
    )]


@api_callable
@reconnect_db_on_error
def get_components_by_host(host_id):
    """Returns a list of components which have the given host added"""
    # TODO: do some logging?
    return [model_to_dict(pc) for pc in models.ProcessComponent.objects.filter(
        hosts__id=host_id
    )]


@api_callable
@reconnect_db_on_error
def get_components_by_service(service_id):
    """Returns a list of components which have the given service added"""
    # TODO: do some logging?
    return [model_to_dict(pc) for pc in models.ProcessComponent.objects.filter(
        services__id=service_id
    )]
