# -*- coding: UTF-8 -*-
# pylint: disable=missing-docstring,too-few-public-methods,no-init,old-style-class,no-member

"""Models for the Business Process Engine (BPE)"""

from django.db import models

# Project imports
from symonitoring_model     import models as models_sm
from symonitoring_rpcserver import livestatus


class ProcessApplication(models.Model):
    """Model for a BPE Application"""

    # Possible values of the state method
    STATE_OK       = 0
    STATE_WARN     = 1
    STATE_CRITICAL = 2
    STATE_UNKNOWN  = 3

    name          = models.CharField(max_length=255, unique=True)
    customer      = models.ForeignKey(models_sm.Customer)
    description   = models.CharField(max_length=255, blank=True, null=True)
    min_available = models.PositiveSmallIntegerField(
        help_text='If less than this number of associated components are '
                  'down, the application is considered to be down.'
    )

    @property
    def state(self):
        current_state   = ProcessApplication.STATE_UNKNOWN
        available_count = 0

        for component in self.processcomponent_set.filter(parent__isnull=True):
            if component.state in (
                ProcessComponent.STATE_OK,
                ProcessComponent.STATE_WARN
            ):
                available_count += 1

        if available_count < self.min_available:
            current_state = ProcessApplication.STATE_CRITICAL
        else:
            current_state = ProcessApplication.STATE_OK

        return current_state

    def superapplications(self, seen=None):
        apps = ProcessApplication.objects.filter(
            processcomponent__applications = self
        )

        seen = [] if seen is None else seen

        if self.pk in seen:
            raise RuntimeError('Applications are circular!')

        ids = [self.pk]
        for app in apps:
            ids.extend(app.superapplications(seen + [self.pk]))

        return ids

    def subapplications(self, seen=None):
        apps = ProcessApplication.objects.filter(
            referencing_components__application = self
        )

        seen = [] if seen is None else seen

        if self.pk in seen:
            raise RuntimeError('Applications are circular!')

        ids = [self.pk]
        for app in apps:
            ids.extend(app.subapplications(seen + [self.pk]))
        return ids


class ProcessComponent(models.Model):
    """Model for a BPE Component"""

    # Possible values of the state method
    STATE_OK       = 0
    STATE_WARN     = 1
    STATE_CRITICAL = 2
    STATE_UNKNOWN  = 3

    application = models.ForeignKey('ProcessApplication')
    parent = models.ForeignKey('self', null=True)
    name = models.CharField(max_length=255)
    description   = models.CharField(max_length=255, blank=True, null=True)
    min_available = models.PositiveSmallIntegerField(
        help_text='If less than this number of associated hosts/services are '
                  'down, the component is considered to be down. Useful for '
                  'cluster and application monitoring.'
    )

    services = models.ManyToManyField(
        'symonitoring_model.Hostserviceinstance',
        through='ProcessService',
    )

    hosts = models.ManyToManyField(
        'symonitoring_model.Host',
        through='ProcessHost',
    )

    applications = models.ManyToManyField(
        'ProcessApplication',
        related_name='referencing_components'
    )

    class Meta:
        unique_together = (('application', 'name'),)

    @property
    def state(self):
        current_state   = ProcessComponent.STATE_UNKNOWN
        available_count = 0

        for host in self.processhost_set.all():
            if host.state == ProcessHost.STATE_UP:
                available_count += 1

        if available_count >= self.min_available:
            return ProcessComponent.STATE_OK

        for service in self.processservice_set.all():
            if service.state in (
                ProcessService.STATE_OK,
                ProcessService.STATE_WARN
            ):
                available_count += 1

        if available_count >= self.min_available:
            return ProcessComponent.STATE_OK

        for application in self.applications.all():
            if application.state in (
                ProcessApplication.STATE_OK,
                ProcessApplication.STATE_WARN
            ):
                available_count += 1

        if available_count >= self.min_available:
            return ProcessComponent.STATE_OK

        for subcomponent in self.processcomponent_set.all():
            if subcomponent.state == ProcessComponent.STATE_OK:
                available_count += 1

        if available_count >= self.min_available:
            return ProcessComponent.STATE_OK
        else:
            return ProcessComponent.STATE_CRITICAL

        return current_state


class ProcessHost(models.Model):
    """Model for a Host associated to a BPE Component"""

    # Possible values of the state property
    STATE_UP          = 0
    STATE_DOWN        = 1
    STATE_UNREACHABLE = 2
    STATE_UNKNOWN     = 3

    # The model properties
    component = models.ForeignKey('ProcessComponent')
    host = models.ForeignKey('symonitoring_model.Host')

    def _icinga_state(self):
        """Return the current host state according to Icinga"""
        try:
            return livestatus.get_host_state(self.host)
        except livestatus.NotFound:
            return self.STATE_UNKNOWN
    state = property(_icinga_state)


class ProcessService(models.Model):
    """Model for a Service instance associated to a BPE Component"""

    # Possible values of the state property
    STATE_OK       = 0
    STATE_WARN     = 1
    STATE_CRITICAL = 2
    STATE_UNKNOWN  = 3

    # The model properties
    component = models.ForeignKey('ProcessComponent')
    service = models.ForeignKey('symonitoring_model.Hostserviceinstance')

    def _icinga_state(self):
        try:
            return livestatus.get_service_state(self.service)
        except livestatus.NotFound:
            return self.STATE_UNKNOWN
        return var
    state = property(_icinga_state)
