# -*- coding: UTF-8 -*-
# pylint: disable=missing-docstring

"""Tests for symonitoring_bpe"""

import mock

from django.test import TestCase

import api
from . import models
from symonitoring_model import models as models_sm

from unittest import skip


class APITests(TestCase):
    """Test symonitoring_bpe.api.API"""
    fixtures = ['test_data']
    maxDiff = None

    def test_return_whole_application_list(self):
        """Test getting the application list from test_data fixture"""
        application_list = api.get_application_list()
        self.assertIn('applications', application_list)
        self.assertEqual(
            len(application_list['applications']),
            3
        )
        self.assertEqual(
            application_list['applications'][0]['id'],
            1
        )
        self.assertEqual(
            application_list['applications'][1]['min_available'],
            3
        )
        self.assertEqual(
            application_list['applications'][2]['id'],
            3
        )
        self.assertEqual(
            application_list['applications'][2]['name'],
            'app3'
        )
        self.assertListEqual(
            sorted(application_list['applications'][0].keys()),
            sorted(['customer', u'id', 'description', 'min_available', 'name'])
        )

    def test_empty_application_list(self):
        models.ProcessApplication.objects.all().delete()  # pylint: disable=no-member
        application_list = api.get_application_list()
        self.assertIn('applications', application_list)
        self.assertEqual(
            len(application_list['applications']),
            0
        )

    @skip("this test has to be changed for the new version of the API")
    def test_get_application(self):
        """Test getting app1 from the test_data fixture"""
        patch_socket = mock.patch('symonitoring_bpe.livestatus.Socket')
        patch_host = mock.patch(
            'symonitoring_bpe.models.ProcessHost.state',
            new_callable=mock.PropertyMock
        )
        patch_service = mock.patch(
            'symonitoring_bpe.models.ProcessService.state',
            new_callable=mock.PropertyMock
        )
        with patch_socket, patch_host as m_host, patch_service as m_service:
            m_host.return_value = 0
            m_service.return_value = 0
            application = api.get_application(1)
            self.assertListEqual(
                sorted(application.keys()),
                sorted([
                    'id',
                    'name',
                    'customer',
                    'description',
                    'min_available',
                    'components',
                    'state'
                ])
            )
            # application
            self.assertEqual(application['id'], 1)
            self.assertEqual(application['min_available'], 1)
            self.assertEqual(application['name'], 'app1')
            self.assertEqual(application['state'], 0)
            self.assertEqual(len(application['components']), 2)
            # component
            component = application['components'][0]
            self.assertEqual(
                sorted(component.keys()),
                sorted([
                    'id',
                    'name',
                    'application',
                    'description',
                    'components',
                    'state',
                    'min_available',
                    'parent',
                    'applications',
                    'hosts',
                    'services'
                ])
            )
            self.assertEqual(component['id'], 1)
            self.assertEqual(component['name'], 'app1comproot')
            self.assertEqual(component['application'], 1)
            self.assertEqual(component['state'], 0)
            self.assertIsNone(component['parent'])
            self.assertEqual(len(component['applications']), 0)
            self.assertEqual(len(component['hosts']), 0)
            self.assertEqual(len(component['services']), 0)
            # child1
            child1 = component['components'][0]
            self.assertEqual(len(child1['applications']), 0)
            self.assertEqual(len(child1['hosts']), 2)
            self.assertEqual(len(child1['services']), 0)
            self.assertEqual(child1['parent'], 1)
            host = child1['hosts'][0]
            self.assertEqual(
                sorted(host.keys()),
                sorted([
                    'id',
                    'name',
                    'state',
                    'type',
                ])
            )
            self.assertEqual(host['state'], 0)
            self.assertEqual(host['type'], 'host')
            # child2
            child2 = component['components'][1]
            self.assertEqual(len(child2['components']), 0)
            self.assertEqual(len(child2['hosts']), 0)
            self.assertEqual(len(child2['services']), 3)
            service = child2['services'][1]
            self.assertEqual(
                sorted(service.keys()),
                sorted([
                    'id',
                    'name',
                    'state',
                    'type',
                    'host_id',
                    'host_name',
                ])
            )
            self.assertEqual(service['state'], 0)
            self.assertEqual(service['type'], 'service')

    def test_get_invalid_application(self):
        with self.assertRaises(IndexError):
            api.get_application(0)

    def test_set_application_min_available(self):
        application = models.ProcessApplication.objects.first()
        old_min_available = application.min_available
        api.set_application_min_available(application.pk, 30)
        application.refresh_from_db()
        self.assertNotEqual(application.min_available, old_min_available)
        self.assertEqual(application.min_available, 30)

    def test_set_component_min_available(self):
        component = models.ProcessComponent.objects.first()
        old_min_available = component.min_available
        api.set_component_min_available(component.pk, 30)
        component.refresh_from_db()
        self.assertNotEqual(component.min_available, old_min_available)
        self.assertEqual(component.min_available, 30)

    def test_rename_application(self):
        application = models.ProcessApplication.objects.first()
        old_name = application.name
        application_id = application.pk
        api.rename_application(application_id, "new")
        application.refresh_from_db()
        self.assertEqual(application.pk, application_id)
        self.assertNotEqual(application.name, old_name)
        self.assertEqual(application.name, "new")

    @skip("this test has to be changed for the new version of the API")
    def test_create_application(self):
        response = api.create_application(
            "ThisIsAReallyLongApplication Name to test",
            1,
            0,
            "ThisIsAReallyShortApplicationDescription to test"
        )
        new_application = models.ProcessApplication.objects.get(
            pk=response['id'])
        self.assertEqual(new_application.min_available, 0)
        self.assertEqual(
            new_application.name,
            "ThisIsAReallyLongApplication Name to test"
        )

    def test_delete_application(self):
        api.create_application("new", 1, 0, "description")
        new_application = models.ProcessApplication.objects.last()
        api.delete_application(new_application.pk)
        self.assertNotIn(
            new_application,
            models.ProcessApplication.objects.all()
        )

    def test_rename_component(self):
        component = models.ProcessComponent.objects.first()
        old_name = component.name
        component_id = component.pk
        api.rename_component(component_id, "new name")
        component.refresh_from_db()
        self.assertEqual(component.pk, component_id)
        self.assertNotEqual(component.name, old_name)
        self.assertEqual(component.name, "new name")

    @skip("this test has to be changed for the new version of the API")
    def test_create_component(self):
        api.create_application("new", 1, 0, "description")
        application = models.ProcessApplication.objects.first()
        response = api.create_component(
            "supercomponent",
            0,
            application.pk,
            "supercomponent",
            "description"
        )
        new_component = models.ProcessComponent.objects.get(
            pk=response['id']
        )
        self.assertEqual(new_component.min_available, 0)
        self.assertEqual(new_component.application, application)
        self.assertEqual(
            new_component.name,
            "supercomponent"
        )

    @skip("this test has to be changed for the new version of the API")
    def test_delete_component(self):
        api.create_application("new", 1, 0, "description")
        application = models.ProcessApplication.objects.first()
        api.create_component(
            "supercomponent",
            0,
            application.pk,
            "supercomponent",
            "description"
        )
        component = models.ProcessComponent.objects.last()
        api.delete_component(component.pk)
        self.assertNotIn(component, models.ProcessComponent.objects.all())

    def test_add_service_to_component(self):
        component = models.ProcessComponent.objects.last()
        instance = models_sm.Hostserviceinstance.objects.last()
        response = api.add_service_to_component(
            component.pk,
            instance.pk
        )
        service = models.ProcessService.objects.get(pk=response['id'])
        self.assertEqual(component, service.component)
        self.assertEqual(instance, service.service)

    def test_add_host_to_component(self):
        component = models.ProcessComponent.objects.last()
        host = models_sm.Host.objects.last()
        response = api.add_host_to_component(
            component.pk,
            host.pk
        )
        process_host = models.ProcessHost.objects.get(pk=response['id'])
        self.assertEqual(component, process_host.component)
        self.assertEqual(host, process_host.host)

    @skip("this test has to be changed for the new version of the API")
    def test_remove_service_from_component(self):
        service = models.ProcessService.objects.last()
        # TODO: what component.pk do we pass here?
        api.remove_service_from_component(service.pk, 1)
        self.assertNotIn(service, models.ProcessService.objects.all())

    @skip("this test has to be changed for the new version of the API")
    def test_remove_host_from_component(self):
        host = models.ProcessHost.objects.last()
        # TODO: what component.pk do we pass here?
        api.remove_host_from_component(host.pk, 1)
        self.assertNotIn(host, models.ProcessHost.objects.all())

    def test_get_host_list(self):
        patch_socket = mock.patch('symonitoring_bpe.livestatus.Socket')
        patch_host = mock.patch(
            'symonitoring_bpe.livestatus.get_host_state'
        )
        with patch_socket, patch_host as m_host:
            m_host.return_value = 0
            hosts = api.get_host_list_by_ldap_group(
                'cn=admin,ou=groups,dc=symon,dc=vm'
            )
        self.assertEqual(len(hosts), 2)
        self.assertEqual(hosts[0]['state'], 0)
        self.assertEqual(
            sorted(hosts[0].keys()),
            sorted([
                'id',
                'name',
                'description',
                'ipv4',
                'ipv6',
                'state',
                'type',
                'sla',
                'customer',
                'os',
                'os_version',
                'isenabled',
            ])
        )

    def test_get_service_list(self):
        patch_socket = mock.patch('symonitoring_bpe.livestatus.Socket')
        patch_service = mock.patch(
            'symonitoring_bpe.livestatus.get_service_state'
        )
        with patch_socket, patch_service as m_service:
            m_service.return_value = 0
            services = api.get_service_list_by_host(2)
        self.assertEqual(len(services), 2)
        self.assertEqual(services[0]['state'], 0)
        self.assertEqual(
            sorted(services[0].keys()),
            sorted(['id', 'name', 'state', 'isenabled', 'sla', 'host'])
        )

    def test_set_parent_component(self):
        component = models.ProcessComponent.objects.get(pk=4)
        parent = models.ProcessComponent.objects.get(pk=3)
        api.set_parent_component(component.pk, parent.pk)
        component.refresh_from_db()
        self.assertEqual(component.parent, parent)

    def test_unset_parent_component(self):
        component = models.ProcessComponent.objects.get(pk=3)
        api.unset_parent_component(component.pk)
        component.refresh_from_db()
        self.assertEqual(component.parent, None)
