# -*- coding: UTF-8 -*-
# pylint:disable=missing-docstring,import-error,bad-whitespace

"""API to access the RE (reporting engine) over RPC"""

# System imports
from copy                               import deepcopy
from datetime                           import datetime, timedelta
from django.db.models                   import Q
from django.db.models.expressions       import RawSQL

# Project imports
from symonitoring_model                 import models as sm_models
from symonitoring_reporting             import models as sr_models
from symonitoring_bpe                   import models as sb_models
from symonitoring_rpcserver.apiregistry import api_callable


# TODO: can these be put in a better place?
# keys to iterate through when aggregating data
_KEYS = (
    'unavailable',
    'unavailable_wo_slatimes',
    'unavailable_wo_downtimes',
    'unavailable_wo_downtimes_slatimes'
)

_KEYS_SLA = (
    'unavailable_wo_slatimes',
    'unavailable_wo_downtimes_slatimes'
)

_STATE_LIST = [
    sm_models.Alert.STATE_CRITICAL,
    sm_models.Alert.STATE_DOWN,
    sm_models.Alert.STATE_UNREACHABLE
]


# _*_to_dict allow to serialize objects the way we need them to be serialized
def _alert_to_dict(alert):
    return {
        'id': alert.id,
        'occurred':     _datetime_to_dict(alert.datetimeoccured),
        'recovery':     _datetime_to_dict(alert.datetimerecovery),
        'acknowledged': alert.datetimeacknowledged and _datetime_to_dict(
            alert.datetimeacknowledged
        ),
        'state':        alert.state,
        'output':       alert.output,
        'comment':      alert.comment,
        'problem':      alert.problemid,
        'duration':     (
            alert.datetimerecovery - alert.datetimeoccured
        ).total_seconds(),
        'user_comment': [_alert_comment_to_dict(c) for c in alert.alertcomment_set.order_by('timestamp')],  # noqa
        'is_service':   bool(alert.serviceinstanceid),
        'type':         'alert',
    }


def _alert_comment_to_dict(comment):
    return {
        'id': comment.id,
        'timestamp': _datetime_to_dict(comment.timestamp),
        'comment': comment.comment,
    }


def _application_to_dict(app):
    issues = (
        getattr(app, 'num_states', 0)
    )
    return {
        'id':          app.pk,
        'customer':    app.customer and _customer_to_dict(app.customer),
        'name':        app.name,
        'description': app.description,
        'states':      getattr(app, 'num_states', 0),
        'issues':      issues,
        'type':        'application'
    }


def _customer_to_dict(cust):
    return {
        'id':       cust.pk,
        'name':     cust.name,
        'sla': {
            'id':   cust.slaid.pk,
            'name': cust.slaid.name,
        },
    }


def _date_to_dict(date):
    return {
        'dayofweek':    date.dayofweek,
        'dayofmonth':   date.dayofmonth,
        'dayofyear':    date.dayofyear,
        'week':         date.week,
        'month':        date.month,
        'quarter':      date.quarter,
        'year':         date.year
    }


def _datetime_to_dict(datetime):
    return {
        'day':      datetime.day,
        'month':    datetime.month,
        'year':     datetime.year,
        'hour':     datetime.hour,
        'minute':   datetime.minute,
        'second':   datetime.second,
    }


def _downtime_to_dict(down):
    return {
        'id':        down.id,
        'from':      _datetime_to_dict(down.datetimefrom),
        'to':        _datetime_to_dict(down.datetimeto),
        'duration':  (down.datetimeto - down.datetimefrom).total_seconds(),
        'type':      'downtime',
    }


def _host_to_dict(host):
    issues = (
        getattr(host, 'num_host_alerts',    0)
    )
    return {
        'id':             host.pk,
        'name':           host.hostname,
        'description':    host.description,
        'ipv4':           host.ipv4,
        'ipv6':           host.ipv6,
        'customer':       _customer_to_dict(host.customerid),
        'os': {
            'id':         host.osid.pk,
            'name':       host.osid.name,
        },
        'os_version':     host.osversionid and {
            'id':         host.osversionid.pk,
            'name':       host.osversionid.name,
        } or None,
        'isenabled':      host.isenabled,
        'downtimes':      getattr(host, 'num_downtimes',      0),
        'alerts_host':    getattr(host, 'num_host_alerts',    0),
        'alerts_service': getattr(host, 'num_service_alerts', 0),
        'services':       [],
        'issues':         issues,
        'type':           'host'
    }


def _service_to_dict(serv):
    return {
        'id':     serv.pk,
        'name':   serv.display_name,
        'issues': getattr(serv, 'num_service_alerts', 0),
        'type':   'service'
    }


def _state_to_dict(state):
    return {
        'id':             state.id,
        'state':          state.state,
        'application_id': state.application_id,
        'occurred':       _datetime_to_dict(state.datetimeoccured),
        'recovery':       _datetime_to_dict(state.datetimerecovery),
        'alert':          _alert_to_dict(state.alert),
        'duration':       (
            state.datetimerecovery - state.datetimeoccured
        ).total_seconds(),
        'type':           'state'
    }


def _get_prepared_services_for_host(host_id, start, end):
    services = sm_models.Hostserviceinstance.objects.filter(hostserviceid__hostid_id=host_id).annotate(  # noqa
        num_service_alerts=RawSQL("""
            SELECT
                COUNT(`Alert`.`id`) as `num_service_alerts`
            FROM
                `Alert`
            WHERE 
                `Alert`.`hostid` = %s
                AND
                `Alert`.`state` IN %s
                AND
                `Alert`.`datetimeOccured` <= %s
                AND
                `Alert`.`datetimeRecovery` >= %s
                AND
                `Alert`.`serviceInstanceId` = `HostServiceInstance`.`id`
        """, (
            int(host_id),
            _STATE_LIST,
            str(end),
            str(start)
        )))

    return dict([(service.id, _service_to_dict(service)) for service in services])  # noqa


# api callables
@api_callable
def get_aggregated_facts(host_ids, application_ids, start, end):

    datetime_start = datetime.strptime(start, '%Y-%m-%d')
    datetime_end   = (datetime.strptime(end, '%Y-%m-%d')
                      + timedelta(days=1, seconds=-1))
    # plus 1 second because timerange of one day is 00:00:00 to 23:59:59
    total_seconds = (datetime_end - datetime_start).total_seconds() + 1

    # get the facts for the given arguments
    facts = get_facts(host_ids, application_ids, start, end)

    # initialize the structure used to return data
    # since the querysets are complex, let's store them in a variable first
    # and then convert the single elements to dictionnaries afterwards
    hosts = sm_models.Host.objects.filter(id__in=host_ids).annotate(
        num_host_alerts=RawSQL("""
            SELECT
                COUNT(`Alert`.`id`) as `num_host_alerts`
            FROM
                `Alert`
            WHERE 
                `Alert`.`hostid` = `Host`.`id`
                AND
                `Alert`.`state` IN %s
                AND
                `Alert`.`datetimeOccured` <= %s
                AND
                `Alert`.`datetimeRecovery` >= %s
                AND
                `Alert`.`serviceInstanceId` IS NULL
        """, (
            _STATE_LIST,
            str(datetime_end),
            str(datetime_start)
        )),
        num_service_alerts=RawSQL("""
            SELECT
                COUNT(`Alert`.`id`) as `num_service_alerts`
            FROM
                `Alert`
            WHERE 
                `Alert`.`hostid` = `Host`.`id`
                AND
                `Alert`.`state` IN %s
                AND
                `Alert`.`datetimeOccured` <= %s
                AND
                `Alert`.`datetimeRecovery` >= %s
                AND
                `Alert`.`serviceInstanceId` IS NOT NULL
        """, (
            _STATE_LIST,
            str(datetime_end),
            str(datetime_start)
        )),
        num_downtimes=RawSQL("""
            SELECT
                COUNT(`Downtime`.`id`) as `num_downtimes`
            FROM
                `Downtime`
            WHERE 
                `Downtime`.`hostid` = `Host`.`id`
                AND
                `Downtime`.`datetimeFrom` <= %s
                AND
                `Downtime`.`datetimeTo` >= %s
        """, (
            str(datetime_end),
            str(datetime_start)
        ))
    )
    hosts = dict([(host.id, _host_to_dict(host)) for host in hosts])

    apps = sb_models.ProcessApplication.objects.filter(id__in=application_ids).annotate(  # noqa
        num_states=RawSQL("""
            SELECT
                COUNT(`app_state`.`id`) AS `num_states`
            FROM
                `symonitoring_reporting_applicationstate` AS `app_state`
            WHERE
                `app_state`.`application_id` =
                    `symonitoring_bpe_processapplication`.`id`
                AND
                `app_state`.`datetimeoccured` <= %s
                AND
                `app_state`.`datetimerecovery` >= %s
        """, (
            str(datetime_end),
            str(datetime_start)
        ))
    )
    applications = dict([(app.id, _application_to_dict(app)) for app in apps])

    # add aggregated data fields and the list for the facts to be appended to
    update = {
        'unavailable':                       0,
        'unavailable_wo_slatimes':           0,
        'unavailable_wo_downtimes':          0,
        'unavailable_wo_downtimes_slatimes': 0,
        'total_seconds_sla':                 0,
        'facts':                             [],
    }
    for host in hosts.values():
        host.update(deepcopy(update))

        host['services'] = _get_prepared_services_for_host(
            host['id'], datetime_start, datetime_end)

        for service in host['services'].values():
            service.update(deepcopy(update))

    for app in applications.values():
        app.update(deepcopy(update))

    # aggregate the facts for the hosts
    for fact in facts['hosts']:

        # get the host of the fact
        host = fact['host']

        # aggregate data and
        # append the fact's data to the list of facts
        # but remove the information which is already stored
        if fact['serviceinstance']:
            for key in _KEYS:
                hosts[host['id']]['services'][
                    fact['serviceinstance']['id']][key] += fact[key]

            hosts[host['id']]['services'][
                fact['serviceinstance']['id']]['facts'].append(fact)

            hosts[host['id']]['services'][
                fact['serviceinstance']['id']
            ]['total_seconds_sla'] += fact['total_seconds_sla']
        else:
            for key in _KEYS:
                hosts[host['id']][key] += fact[key]

            hosts[host['id']]['facts'].append(fact)
            hosts[host['id']]['total_seconds_sla'] += fact['total_seconds_sla']

        fact.pop('host')
        fact.pop('customer')
        fact.pop('serviceinstance')

    # aggregate the facts for applications
    for fact in facts['applications']:

        # get the application of the fact
        application = fact['application']

        # aggregate data
        for key in _KEYS:
            applications[application['id']][key] += fact[key]

        # append the fact to the list of facts
        fact.pop('application')
        applications[application['id']]['facts'].append(fact)

    # normalize data
    for host in hosts.values():
        # set the available field
        host['available'] = total_seconds - host['unavailable']
        host['available_percent'] = host['available'] / total_seconds * 100
        # renormalize the rest of the fields
        for key in _KEYS:
            if key in _KEYS_SLA:
                if host['total_seconds_sla'] > 0:
                    host['%s_percent' % key] = float(host[key]) / host['total_seconds_sla'] * 100  # noqa
                else:
                    host['%s_percent' % key] = 0
            else:
                host['%s_percent' % key] = host[key] / total_seconds * 100

        # do the same for the services
        for service in host['services'].values():
            # set the available field
            service['available'] = total_seconds - service['unavailable']
            service['available_percent'] = service['available'] / total_seconds * 100  # noqa
            # renormalize the rest of the fields
            for key in _KEYS:
                if key in _KEYS_SLA:
                    if service['total_seconds_sla'] > 0:
                        service['%s_percent' % key] = float(service[key]) / service['total_seconds_sla'] * 100  # noqa
                    else:
                        service['%s_percent' % key] = 0
                else:
                    service['%s_percent' % key] = service[key] / total_seconds * 100  # noqa

    for app in applications.values():
        # set the available field
        app['available'] = total_seconds - app['unavailable']
        app['available_percent'] = app['available'] / total_seconds * 100
        # renormalize the rest of the fields
        for key in _KEYS:
            app['%s_percent' % key] = app[key] / total_seconds * 100

    return {
        'hosts':         hosts.values(),
        'applications':  applications.values(),
        'total_seconds': total_seconds,
    }


def _parse_start_end_date(start, end):
    startdate  = datetime.strptime(start, '%Y-%m-%d')
    enddate    = (datetime.strptime(end, '%Y-%m-%d')
                  + timedelta(days=1, seconds=-1))

    return (startdate, enddate)


@api_callable
def get_alerts_for_host(host_id, start, end):
    startdate, enddate  = _parse_start_end_date(start, end)

    return [_alert_to_dict(a) for a in sm_models.Alert.objects.filter(
        Q(hostid=host_id),
        Q(serviceinstanceid__isnull=True),
        Q(state__in=_STATE_LIST),
        Q(datetimeoccured__lte=enddate) & Q(datetimerecovery__gte=startdate)
    )]


@api_callable
def get_alerts_for_host_inc_services(host_id, start, end):
    startdate, enddate  = _parse_start_end_date(start, end)

    return [_alert_to_dict(a) for a in sm_models.Alert.objects.filter(
        Q(hostid=host_id),
        Q(state__in=_STATE_LIST),
        Q(datetimeoccured__lte=enddate) & Q(datetimerecovery__gte=startdate)
    )]


@api_callable
def get_alerts_for_service(id, start, end):
    startdate, enddate  = _parse_start_end_date(start, end)

    return [_alert_to_dict(a) for a in sm_models.Alert.objects.filter(
        Q(serviceinstanceid=id),
        Q(state__in=_STATE_LIST),
        Q(datetimeoccured__lte=enddate) & Q(datetimerecovery__gte=startdate)
    )]


@api_callable
def get_states_for_application(application_id, start, end):
    startdate, enddate  = _parse_start_end_date(start, end)

    return [_state_to_dict(s) for s in sr_models.ApplicationState.objects.filter(  # noqa
        Q(application_id=application_id) &
        Q(datetimeoccured__lte=enddate) &
        Q(datetimerecovery__gte=startdate)
    )]


@api_callable
def get_downtimes_for_host(host_id, start, end):
    startdate, enddate  = _parse_start_end_date(start, end)

    return [_downtime_to_dict(d) for d in sm_models.Downtime.objects.filter(
        Q(hostid=host_id),
        Q(datetimefrom__lte=enddate) & Q(datetimeto__gte=startdate)
    )]


@api_callable
def get_downtimes_for_application(application_id, start, end):
    startdate, enddate  = _parse_start_end_date(start, end)

    return [_downtime_to_dict(d) for d in sm_models.Downtime.objects.filter(
        Q(
            hostid__processcomponent__application=application_id,
            serviceinstanceid__isnull=True
        ) | Q(
            serviceinstanceid__processcomponent__application=application_id
        ),
        Q(datetimefrom__lte=enddate) & Q(datetimeto__gte=startdate)
    )]


@api_callable
def get_facts(host_ids, application_ids, start, end):
    return {
        'hosts': get_facts_for_hosts(host_ids, start, end),
        'applications': get_facts_for_applications(application_ids, start, end)
    }


@api_callable
def get_facts_for_hosts(host_ids, start, end):
    return [{
        'date':                              _date_to_dict(f.date),
        'unavailable':                       f.unavailable,
        'unavailable_wo_slatimes':           f.unavailable_wo_slatimes,
        'unavailable_wo_downtimes':          f.unavailable_wo_downtimes,
        'unavailable_wo_downtimes_slatimes': f.unavailable_wo_downtimes_slatimes,  # noqa
        'total_seconds_sla':                 f.total_seconds_sla,
        'customer':                          _customer_to_dict(f.customer),
        'host':                              _host_to_dict(f.host),
        'serviceinstance': f.serviceinstance and _service_to_dict(f.serviceinstance) or None,  # noqa
        'type':                              'host'
    } for f in sr_models.BiFactavailability.objects.filter(
        host__in=host_ids,
        date__date__range=(
            datetime.strptime(start, '%Y-%m-%d').date(),
            datetime.strptime(end, '%Y-%m-%d').date()
        )
    )]


@api_callable
def get_facts_for_applications(application_ids, start, end):
    return [{
        'date':                              _date_to_dict(f.date),
        'unavailable':                       f.unavailable,
        'unavailable_wo_slatimes':           f.unavailable_wo_slatimes,
        'unavailable_wo_downtimes':          f.unavailable_wo_downtimes,
        'unavailable_wo_downtimes_slatimes': f.unavailable_wo_downtimes_slatimes,  # noqa
        'customer':                          _customer_to_dict(f.customer),
        'application':                       _application_to_dict(f.application),  # noqa
        'type':                              'application'
    } for f in sr_models.BiAppFactavailability.objects.filter(
        application__in=application_ids,
        date__date__range=(
            datetime.strptime(start, '%Y-%m-%d').date(),
            datetime.strptime(end, '%Y-%m-%d').date()
        )
    )]


@api_callable
def get_incidents_for_application(id, start, end):
    return {
        'states': get_states_for_application(id, start, end),
        'downtimes': get_downtimes_for_application(id, start, end),
    }


@api_callable
def get_incidents_for_host(id, start, end):
    return {
        'alerts': get_alerts_for_host(id, start, end),
        'downtimes': get_downtimes_for_host(id, start, end)
    }


@api_callable
def get_incidents_for_service(id, start, end):
    return {
        'alerts': get_alerts_for_service(id, start, end)
    }
