import datetime


def process(start, end):
    """
    Yields duration (and date) of time between beginning and end for each day.
    If the time spans two or more dates, then a corresponding amount of
    duration and date tuples will be returned.
    Beginning and end must be in order.
    Accepts only two datetime objects.

    day_start : beginning of a day
    day_end   : ending of a day
    start     : beginning of an alert
    end       : ending of an alert
    """
    day_start = datetime.datetime(
        start.year, start.month, start.day,
        0, 0, 0, tzinfo=start.tzinfo)

    # while our current day is still before or on the same
    # day as the end day we calculate:
    while day_start < end:
        day_end = day_start + datetime.timedelta(days=1)

        # we take min of day_end and end, when we reach our end point
        # and we take the max of day_start or start
        # this is mainly useful, when an outtage overspans
        # multiple days
        duration = min(day_end, end) - max(day_start, start)

        yield (duration.total_seconds(), day_start.date())
        day_start = day_end


def process_within_multiple_slatimes(begin, end, timeframes, downtimes=None):
    """
    Yields duration and date tuple of a timespan between begin and end
    and within timeframes provided.
    Timeframes are expected to be a dict, ordered by the datetime.day()
    equivalent of their weekday.
    E.g. monday is 0, sunday is 6.
    One timeframe is expected to be a tuple of datetime.time objects,
    one being timefrom and the other timeto.

    Timeframes may not overlap!
    """
    assert begin <= end
    #  day is our current day we want to calculate the unavailable time for
    day = datetime.datetime(
        begin.year, begin.month, begin.day,
        0, 0, 0, tzinfo=begin.tzinfo)

    # while our current day is still before or on the same
    # day as the end day we calculate:
    while day < end:
        day_end = day + datetime.timedelta(days=1)
        unavailable = 0

        #  because when there may not be a timeframe for each day
        #  we will continue if there is none
        tmp_timeframes = timeframes.get(day.date().weekday())
        if (tmp_timeframes) is None:
            day = day_end
            continue

        # counter for the max possible sla seconds
        total_seconds_sla = 0

        #  we could have multiple timeframes for this day of the week
        #  so we will loop through each timeframe, that matches this week day
        for timefrom, timeto in tmp_timeframes:
            #  make times timezone aware
            timefrom = timefrom.replace(tzinfo=day.tzinfo)
            timeto   = timeto.replace(tzinfo=day.tzinfo)

            #  if timeframes are in the wrong order we should
            #  change them
            if timeto < timefrom:
                tmp_time = timeto
                timeto = timefrom
                timefrom = tmp_time

            #  make datetime objects for this frame on this day
            datetime_from = datetime.datetime.combine(day.date(), timefrom)
            datetime_to = datetime.datetime.combine(day.date(), timeto)

            # count sla seconds
            sla_timeframe_diff = datetime_to - datetime_from
            total_seconds_sla += sla_timeframe_diff.total_seconds()

            #  if timeframe ends before the outtage begins we should move along
            if datetime_to < begin or datetime_from > end:
                continue
            #  now find out, how long the host/service was unavailable from
            #  this timeframe
            else:
                final_from   = max(datetime_from, begin)
                final_to     = min(datetime_to, end)
                timespan     =  final_to - final_from
                unavailable += timespan.total_seconds()

                if downtimes is not None:
                    for unav_downtime, date in process_wo_downtimes(
                        final_from,
                        final_to,
                        downtimes
                    ):
                        unavailable -= unav_downtime

        #  beware of dragons: if we have an issue which reaches from one day
        #  into an other and the sla timeframe is set to the time 23:59, the
        #  customer means actually until midnight (because he can't input
        #  24:00)! Because of this, we would be missing 60 seconds, so "simply"
        #  add them here..
        if [e for e in tmp_timeframes if e[1].hour == 23 and e[1].minute == 59]:  # noqa
            total_seconds_sla += 60

            if begin.day != day.day:
                unavailable += 60

        #  and yield our result
        yield(unavailable, day.date(), total_seconds_sla)
        #  and set the day to the next day
        day = day_end


def process_wo_downtimes(begin, end, downtimes):
    """
    Yields duration and date of time between beginning and end for each day
    exluding any occuring downtimes on similar dates.
    If beginning and end span two or more dates, then a corresponding amount
    of duration and date tuples will be returned.
    Expects a list of datetime.datetime tuples as downtimes

    Yields tuples of seconds(int), dates(datetime.date)
    """
    assert begin <= end
    # day, end, day_end, dt_start, dt_end, dt_date are all datetime objects
    #  day is our current day we want to calculate the unavailable time for
    day = datetime.datetime(
        begin.year, begin.month, begin.day,
        0, 0, 0, tzinfo=begin.tzinfo)

    # while our current day is still before or on the same
    # day as the end day we calculate:
    while day < end:
        day_end = day + datetime.timedelta(days=1)
        dt_seconds = 0

        # because we can potentially have more than one
        # planned downtime in one day, we must sum up all
        # downtimes during our outtage
        for dt_start, dt_end in downtimes:

            # because downtimes potentially span over multiple days
            # we must use process for calculating the duration of a downtime
            datetimes = process(max(dt_start, begin), min(dt_end, end))
            for dt_duration, dt_date in datetimes:

                # here we do the summing up, but we only sum up the
                # the days that match ours.
                # TODO: This could probably be simplified - not using process()
                if dt_date == day.date():
                    dt_seconds += dt_duration

        # the actual duration of an outtage without the planned down
        # now equals the full duration of the outtage minus the summed
        # up downtimes
        duration = (
            min(day_end, end) -
            max(day, begin) - datetime.timedelta(seconds=dt_seconds))

        yield(duration.total_seconds(), day.date())
        day = day_end
