#!/usr/bin/env python2
# -*- coding: UTF-8 -*-
# vim: autoindent expandtab tabstop=4 sw=4 sts=4 filetype=python

from datetime import datetime, timedelta

from django.core.management.base  import BaseCommand, CommandError
from symonitoring_bpe             import models as sb_models
from symonitoring_model           import models as sm_models
from symonitoring_reporting       import models as sr_models
from symonitoring_reporting       import duration


class Command(BaseCommand):

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self._log_adapter = None
        self.alerts = None

        self.verbose = False

    def add_arguments(self, parser):
        """Deals with the arguments given to the reporting command"""
        parser.add_argument(
            '--datefrom', '-f', type=str,
            help='Start date for processing reports'
        )
        parser.add_argument(
            '--dateto', '-t', type=str,
            help='End date for processing reports'
        )

    def handle(self, *args, **options):
        startdate = None
        enddate   = None

        if options['dateto']:
            enddate = (self.parsedate(options['dateto'])
                       + timedelta(days=1, seconds=-1))
        if options['datefrom']:
            startdate = self.parsedate(options['datefrom'])
        if options['verbosity']:
            self.verbose = True

        # reset unavailable times
        self._log("Reset all computed times within range...")
        self._reset_application_states(startdate, enddate)
        self._reset_unavailbable_times(startdate, enddate)

        # compute unavailable times
        self._log("Fetching relevant alerts...")
        self.alerts = self.load_alerts(startdate, enddate)

        self._log("Applying simulations...")
        self._simulate_application_events(self.alerts)

        self._log("Compute app facts...")
        application_states = self.load_app_events(
            startdate,
            enddate)
        for state in application_states:
            self._compute_unavailable_for_app(state)

        # end computing with creating a batch job
        self._log("Create batch job")
        sr_models.ReportingBatchjob.objects.create(
            timestamp=datetime.now(),
            datefrom=startdate or datetime.now().date(),
            dateto=enddate or datetime.now().date(),
            successful=True
        )

    def _log(self, msg):
        if self.verbose:
            self.stdout.write(msg)

    @staticmethod
    def parsedate(date):
        try:
            return datetime.strptime(date, '%Y-%m-%d')
        except ValueError as err:
            raise CommandError(
                'Invalid date format. Expected '
                'YYYY-MM-DD, got: {}. \n {}'.format(date, err)
            )

    @staticmethod
    def _reset_application_states(startdate=None, enddate=None):
        sr_models.ApplicationState.objects.filter(
            datetimerecovery__gte=startdate,
            datetimeoccured__lte=enddate
        ).delete()

    @staticmethod
    def _reset_unavailbable_times(startdate=None, enddate=None):
        for Model in (
            sr_models.BiFactavailability,
            sr_models.BiAppFactavailability
        ):
            facts = Model.objects.all()
            if startdate:
                facts = facts.filter(date__date__gte=startdate)
            if enddate:
                facts = facts.filter(date__date__lte=enddate)
            facts.update(
                unavailable = 0,
                unavailable_wo_slatimes = 0,
                unavailable_wo_downtimes = 0,
                unavailable_wo_downtimes_slatimes = 0
            )

    @staticmethod
    def load_alerts(startdate, enddate):
        state_list = [
            sm_models.Alert.STATE_CRITICAL,
            sm_models.Alert.STATE_DOWN,
            sm_models.Alert.STATE_UNREACHABLE
        ]
        filtered = sm_models.Alert.objects.filter(
            datetimeoccured__isnull=False,
            datetimerecovery__isnull=False,
            state__in=state_list
        )
        if startdate:
            filtered = filtered.filter(datetimerecovery__gte=startdate)
        if enddate:
            filtered = filtered.filter(datetimeoccured__lte=enddate)

        return filtered

    @staticmethod
    def load_app_events(startdate, enddate):
        filtered = sr_models.ApplicationState.objects.filter(
            datetimeoccured__isnull=False,
            datetimerecovery__isnull=False,
        )
        if startdate:
            filtered = filtered.filter(datetimeoccured__lte=enddate)
        if enddate:
            filtered = filtered.filter(datetimerecovery__gte=startdate)
        return filtered

    @staticmethod
    def get_all_events(alerts):
        def all_events(alerts):
            for alert in alerts:
                yield (alert.datetimeoccured, 'DOWN', alert)
                if alert.datetimerecovery:
                    yield (alert.datetimerecovery, 'UP', alert)
        return sorted(all_events(alerts), key=lambda x: x[0])

    @staticmethod
    def load_downtimes(hostid, datetime_from, datetime_to):
        downtime_list = []
        for downtime in sm_models.Downtime.objects.filter(
            hostid=hostid,
            datetimeto__gte=datetime_from,
            datetimefrom__lte=datetime_to
        ):
            downtime_list.append((
                downtime.datetimefrom,
                downtime.datetimeto
            ))

        return downtime_list

    @classmethod
    def _app_to_dict(cls, app):
        # we assume that at the beginning, the application is totally available
        total_avail = app.processcomponent_set.all().count()
        return {
            'id'          : app.id,
            'min_avail'   : app.min_available,
            'curr_act'    : total_avail,
            'total_avail' : total_avail,
            'event'       : None
        }

    @classmethod
    def _comp_to_dict(cls, comp):
        # we assume that at the beginning, the application is totally available
        total_avail = (
            comp.processcomponent_set.all().count() +
            comp.applications.all().count() +
            comp.hosts.all().count() +
            comp.services.all().count()
        )
        return {
            'id'          : comp.id,
            'min_avail'   : comp.min_available,
            'curr_act'    : total_avail,
            'total_avail' : total_avail
        }

    @classmethod
    def _host_to_dict(cls, host):
        return {
            'id'          : host.id,
            'min_avail'   : 1,
            'curr_act'    : 1,
            'total_avail' : 1,
            'event'       : None
        }

    @classmethod
    def _service_to_dict(cls, service):
        return {
            'id'          : service.id,
            'min_avail'   : 1,
            'curr_act'    : 1,
            'total_avail' : 1,
            'event'       : None
        }

    def _initalize_simulation(self, alerts):

        apps = sb_models.ProcessApplication.objects.all()
        comps = sb_models.ProcessComponent.objects.all()
        hosts = sm_models.Host.objects.filter(
            alert__in=alerts
        )
        services = sm_models.Hostserviceinstance.objects.filter(
            alert__in=alerts
        )

        parents = {}
        objects = {}

        for app in apps:
            objects['app:%d' % app.id] = self._app_to_dict(app)
            parents['app:%d' % app.id] = [
                'comp:%d' % comp.id
                for comp in app.referencing_components.all()
            ]

        for comp in comps:
            objects['comp:%d' % comp.id] = self._comp_to_dict(comp)
            if comp.parent:
                parents['comp:%d' % comp.id] = ['comp:%d' % comp.parent.id]
            else:
                parents['comp:%d' % comp.id] = ['app:%d' % comp.application.id]

        for host in hosts:
            objects['host:%d' % host.id] = self._host_to_dict(host)
            parents['host:%d' % host.id] = [
                'comp:%d' % comp.id
                for comp in host.processcomponent_set.all()
            ]

        for service in services:
            objects['service:%d' % service.id] = self._service_to_dict(service)
            parents['service:%d' % service.id] = [
                'comp:%d' % comp.id
                for comp in service.processcomponent_set.all()
            ]

        return objects, parents

    def _apply_event(self, objects, parents, identifier, direction, alert):

        self._log("  identifier %s" % identifier)
        """Storytellin"""
        # sake of shortness
        obj = objects[identifier]

        # state before applying direction
        state_before = obj['curr_act'] < obj['min_avail']
        obj['curr_act'] += direction
        # state after applying direction
        state_after = obj['curr_act'] < obj['min_avail']
        self._log("   curr_act / min_avail / total_avail   %d / %d / %d" % (
            obj['curr_act'],
            obj['min_avail'],
            obj['total_avail']
        ))

        # if there was a status change, we have to bubble up!
        if state_before != state_after:
            for p in parents[identifier]:
                self._log("   change in status! bubble up to %s" % str(p))
                self._apply_event(objects, parents, p, direction, alert)

        # if the object became unavailable and it is application (ergo 'event'
        # is in the list of keys) and the event object is None, then we have to
        # create an application state
        # ('event' is probably the wrong name for the key)
        if 'event' in obj:
            if obj['curr_act'] < obj['min_avail']  and obj['event'] is None:
                if identifier.startswith('app:'):
                    self._log(
                        "   creating application state for %s" % identifier
                    )
                    obj['event'] = sr_models.ApplicationState.objects.create(
                        alert=alert,
                        application=sb_models.ProcessApplication.objects.get(
                            id=identifier.split(':')[-1]
                        ),
                        datetimeoccured=alert.datetimeoccured,
                        state=1
                    )
                    self._log("  application state created %s" % obj['event'])

                elif identifier.startswith('host:'):
                    self._log(
                        "   opening host downtime for %s" % identifier)

                    obj['event'] = {
                        'hostid'          : obj['id'],
                        'serviceid'       : None,
                        'customerid'      : alert.hostid.customerid.id,
                        'slaid'           : alert.slaid.id,
                        'datetimeoccurred': alert.datetimeoccured,
                        'datetimerecovery': None
                    }

                elif identifier.startswith('service:'):
                    self._log(
                        "   opening service downtime for %s" % identifier)

                    obj['event'] = {
                        'hostid'          : alert.hostid.id,
                        'serviceid'       : obj['id'],
                        'customerid'      : alert.hostid.customerid.id,
                        'slaid'           : alert.slaid.id,
                        'datetimeoccurred': alert.datetimeoccured,
                        'datetimerecovery': None
                    }

            if obj['curr_act'] >= obj['min_avail'] and obj['event']:
                if identifier.startswith('app:'):
                    self._log(
                        " closing  application state for %s" % identifier)
                    obj['event'].datetimerecovery = alert.datetimerecovery
                    obj['event'].save()
                    self._log(
                        " application state closed and saved %s" % obj['event']
                    )
                    obj['event'] = None

                elif identifier.startswith('host:') or identifier.startswith('service:'):  # noqa
                    self._log(
                        " downtime ended, write fact for %s" % identifier)

                    obj['event']['datetimerecovery'] = alert.datetimerecovery
                    self._compute_unavailable(obj['event'])
                    self._compute_unavailable_wo_downtimes(obj['event'])
                    self._compute_unavailable_wi_slatimes(obj['event'])
                    self._compute_unavailable_wi_slatimes(obj['event'], True)

                    self._log(
                        " fact created for %s" % obj['event'])
                    obj['event'] = None

    def _simulate_application_events(self, alerts):

        objects, parents = self._initalize_simulation(alerts)

        # go through all the alerts and check if the app is up or not
        for date, state, alert in self.get_all_events(alerts):

            # is the state up, then the direction is +1, else it's -1
            direction = state == 'UP' and 1 or -1

            self._log("event %s -- state %s -- direction %d" % (
                str(date),
                state,
                direction
            ))
            self._log(" alert from %s to %s" % (
                str(alert.datetimeoccured),
                str(alert.datetimerecovery)
            ))

            if alert.serviceinstanceid:
                self._log("  service h%02d.s%02d" % (
                    alert.hostid.id,
                    alert.serviceinstanceid.id
                ))
                identifier = 'service:%d' % alert.serviceinstanceid.id
            else:
                self._log("  host    h%02d" % alert.hostid.id)
                identifier = 'host:%d' % alert.hostid.id

            self._apply_event(objects, parents, identifier, direction, alert)

    def _compute_unavailable(self, timeframe):
        unav_dates = duration.process(
            timeframe['datetimeoccurred'],
            timeframe['datetimerecovery']
        )

        for unavailable, date in unav_dates:
            bi_date, _ = sr_models.BiDimensiondate.objects.get_or_create(
                date = date,
                dayofweek = date.weekday(),
                dayofmonth = date.day,
                month = date.month,
                year = date.year,
                quarter = (date.month - 1) // 3 + 1,
                week = date.isocalendar()[1],
                dayofyear = date.timetuple().tm_yday
            )
            fact, _ = sr_models.BiFactavailability.objects.get_or_create(
                date = bi_date,
                host_id = timeframe['hostid'],
                customer_id = timeframe['customerid'],
                serviceinstance_id = timeframe['serviceid'],
            )
            fact.unavailable += unavailable
            fact.save()

    def _compute_unavailable_wo_downtimes(self, timeframe):
        downtimes = self.load_downtimes(
            timeframe['hostid'],
            timeframe['datetimeoccurred'],
            timeframe['datetimerecovery'])

        unav_dates = duration.process_wo_downtimes(
            timeframe['datetimeoccurred'],
            timeframe['datetimerecovery'],
            downtimes
        )

        for unavailable_wo_downtimes, date in unav_dates:
            bi_date, _ = sr_models.BiDimensiondate.objects.get_or_create(
                date = date,
                dayofweek = date.weekday(),
                dayofmonth = date.day,
                month = date.month,
                year = date.year,
                quarter = (date.month - 1) // 3 + 1,
                week = date.isocalendar()[1])
            fact, _ = sr_models.BiFactavailability.objects.get_or_create(
                date = bi_date,
                host_id = timeframe['hostid'],
                customer_id = timeframe['customerid'],
                serviceinstance_id = timeframe['serviceid'])
            fact.unavailable_wo_downtimes += unavailable_wo_downtimes
            fact.save()

    def _compute_unavailable_wi_slatimes(self, timeframe, exclude_downtimes=False):  # noqa
        try:
            sla_timeframes = sm_models.Timeframe.objects.filter(
                slaid=timeframe['slaid'])
        except sm_models.Timeframe.DoesNotExist:
            self._log("The wanted SLA (slaid %d) does no longer exist!" % timeframe['slaid'])  # noqa
            self._log("The alert from host %d seems to be outdated.." % timeframe['hostid'])  # noqa

        #  Add all timeframes into week
        merged_sla_timeframes = {}
        day_count = 0
        for day in ("monday",
                    "tuesday",
                    "wednesday",
                    "thursday",
                    "friday",
                    "saturday",
                    "sunday"):
            #  We want to merge any overlapping timeframes to avoid
            #  duplicate times.
            #  By also ordering them, we can simplify our merge.
            sla_day_timeframes = sla_timeframes.filter(
                **{day: True}).order_by('timefrom')

            if not len(sla_day_timeframes):
                merged_sla_timeframes[day_count] = []
                day_count += 1
                continue

            merged_sla_timeframes[day_count] = [(tf.timefrom, tf.timeto) for tf in self._timeframe_merge(  # noqa
                sla_day_timeframes)]
            day_count += 1

        downtimes = None
        if exclude_downtimes:
            downtimes = self.load_downtimes(
                timeframe['hostid'],
                timeframe['datetimeoccurred'],
                timeframe['datetimerecovery'])

        dates = duration.process_within_multiple_slatimes(
            timeframe['datetimeoccurred'],
            timeframe['datetimerecovery'],
            merged_sla_timeframes,
            downtimes)

        for unavailable, date, total_seconds_sla in dates:
            bi_date, _ = sr_models.BiDimensiondate.objects.get_or_create(
                date = date,
                dayofweek = date.weekday(),
                dayofmonth = date.day,
                month = date.month,
                year = date.year,
                quarter = (date.month - 1) // 3 + 1,
                week = date.isocalendar()[1]
            )
            fact, _ = sr_models.BiFactavailability.objects.get_or_create(
                date = bi_date,
                host_id = timeframe['hostid'],
                customer_id = timeframe['customerid'],
                serviceinstance_id = timeframe['serviceid']
            )
            if exclude_downtimes:
                fact.unavailable_wo_downtimes_slatimes += unavailable
            else:
                fact.unavailable_wo_slatimes += unavailable

            fact.total_seconds_sla = total_seconds_sla
            fact.save()

    def _compute_unavailable_for_app(self, state):
        unav_dates = duration.process(
            state.datetimeoccured,
            state.datetimerecovery
        )

        for unavailable, date in unav_dates:
            bi_date, _ = sr_models.BiDimensiondate.objects.get_or_create(
                date = date,
                dayofweek = date.weekday(),  # yes, this is a method
                dayofmonth = date.day,
                month = date.month,
                year = date.year,
                quarter = (date.month - 1) // 3 + 1,
                week = date.isocalendar()[1],
                dayofyear = date.timetuple().tm_yday
            )

            fact, _ = sr_models.BiAppFactavailability.objects.get_or_create(
                application = state.application,
                date = bi_date,
                customer = state.alert.hostid.customerid,
            )
            fact.unavailable += unavailable
            fact.save()

    def _timeframe_merge(self, timeframes):
        """
        Merges any overlapping timeframes and returns
        them as generator.

        Timeframes must be ordered by date!
        """
        if len(timeframes) > 1:
            #  If first two timeframes overlap
            if self._do_timeframes_overlap(*timeframes[:2]):
                # As timeframes are ordered, we can merge them by setting the
                # maximum timeto for the first timeframe and discard the second
                timeframes[0].timeto = max(
                    timeframes[0].timeto, timeframes[1].timeto)
                #  Recursion with remaining timeframes
                for tf in self._timeframe_merge([timeframes[0]] + timeframes[2:]):  # noqa
                    yield tf
            else:
                #  Timeframe cannot be merge, so we yield it
                yield timeframes[0]
                #  Recursion with remaining timeframes
                for tf in self._timeframe_merge(timeframes[1:]):
                    yield tf

        #  Trivial case - there are no timeframes to merge
        else:
            yield timeframes[0]

    def _do_timeframes_overlap(self, frame1, frame2):
        """
        Return true if frames overlap, false otherwise
        """
        #  Order frames
        if frame1.timefrom > frame2.timefrom:
            tempframe = frame1
            frame1 = frame2
            frame2 = tempframe
        return frame1.timeto >= frame2.timefrom
