
import datetime

from symonitoring_reporting import duration

from django.test import TestCase

class DurationTest(TestCase):
    def test_single_day():
        begin = datetime.datetime(2015, 12, 6, 18, 0)
        end   = datetime.datetime(2015, 12, 6, 19, 0)

        duration_and_days = list(duration.process(begin, end))
        assert len(duration_and_days) == 1

        dur, date = duration_and_days[0]

        assert dur == 3600
        assert date == datetime.date(2015, 12, 6)


    def test_two_days():
        begin = datetime.datetime(2015, 12, 6, 18, 0)
        end   = datetime.datetime(2015, 12, 7, 19, 0)

        duration_and_days = list(duration.process(begin, end))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 12, 6)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 19 * 3600
        assert date == datetime.date(2015, 12, 7)


    def test_three_days():
        begin = datetime.datetime(2015, 12, 6, 18, 0)
        end   = datetime.datetime(2015, 12, 8, 20, 0)

        duration_and_days = list(duration.process(begin, end))
        assert len(duration_and_days) == 3

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 12, 6)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 24 * 3600
        assert date == datetime.date(2015, 12, 7)

        # Day 2
        dur, date = duration_and_days[2]
        assert dur == 20 * 3600
        assert date == datetime.date(2015, 12, 8)


    def test_leap_hour():
        return  # test does not work as of yet
        begin = datetime.datetime(2015, 10, 24, 18, 0)
        end   = datetime.datetime(2015, 10, 25, 20, 0)

        duration_and_days = list(duration.process(begin, end))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 19 * 3600  # 1 hour less than normal - clocks should go backwards 1h
        assert date == datetime.date(2015, 10, 25)


    def test_single_day_excluding_sla():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 24, 23, 0)
        sla_begin = datetime.time(8)
        sla_end   = datetime.time(19)

        duration_and_days = list(duration.process_excluding_slatimes(begin, end, sla_begin, sla_end))
        assert len(duration_and_days) == 1

        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_two_days_excluding_sla():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)
        sla_begin = datetime.time(8)
        sla_end   = datetime.time(19)

        duration_and_days = list(duration.process_excluding_slatimes(begin, end, sla_begin, sla_end))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 5 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_three_days_excluding_sla():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 26, 13, 0)
        sla_begin = datetime.time(8)
        sla_end   = datetime.time(19)

        duration_and_days = list(duration.process_excluding_slatimes(begin, end, sla_begin, sla_end))
        assert len(duration_and_days) == 3

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 11 * 3600
        assert date == datetime.date(2015, 10, 25)

        # Day 3
        dur, date = duration_and_days[2]
        assert dur == 5 * 3600
        assert date == datetime.date(2015, 10, 26)


    def test_excluding_one_downtime():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 24, 23, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 20, 0)

        downtimes = (dt1_start, dt1_end)
        duration_and_days = list(duration.process_excluding_downtimes(begin, end, [downtimes]))
        assert len(duration_and_days) == 1

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 4 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_excluding_two_downtimes():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 24, 23, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 15, 0)

        dt2_start = datetime.datetime(2015, 10, 24, 20, 0)
        dt2_end   = datetime.datetime(2015, 10, 24, 21, 0)

        downtimes = [(dt1_start, dt1_end), (dt2_start, dt2_end)]
        duration_and_days = list(duration.process_excluding_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 1

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 8 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_two_days_excluding_one_downtime():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 15, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.process_excluding_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 10 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 13 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_two_days_excluding_multiple_day_downtime():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 25, 1, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.process_excluding_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 1 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 12 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_two_days_excluding_multiple_downtimes():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 25, 1, 0)

        dt2_start = datetime.datetime(2015, 10, 25, 12, 0)
        dt2_end   = datetime.datetime(2015, 10, 25, 13, 0)

        downtimes = [(dt1_start, dt1_end), (dt2_start, dt2_end)]
        duration_and_days = list(duration.process_excluding_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 1 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 11 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_downtime_before_outtage():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 11, 0)
        dt1_end   = datetime.datetime(2015, 10, 25, 0, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.process_excluding_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 0 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 13 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_downtime_longer_than_outtage():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 25, 12, 0)
        dt1_end   = datetime.datetime(2015, 10, 25, 15, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.process_excluding_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 11 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 12 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_downtime_overspanning_outtage():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 12, 0)
        dt1_end   = datetime.datetime(2015, 10, 25, 15, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.process_excluding_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 0 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 0 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_one_day_excluding_sla_downtime():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 24, 23, 0)

        sla_begin = datetime.time(8)
        sla_end = datetime.time(19)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 15, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.process_excluding_downtimes_slatimes(begin, end, sla_begin, sla_end, downtimes))
        assert len(duration_and_days) == 1

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 5 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_two_days_excluding_sla_downtime():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 23, 0)

        sla_begin = datetime.time(8)
        sla_end = datetime.time(19)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 15, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.process_excluding_downtimes_slatimes(begin, end, sla_begin, sla_end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 5 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 11 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_two_days_excluding_slatimes_with_long_downtime():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 23, 0)

        sla_begin = datetime.time(8)
        sla_end = datetime.time(19)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 25, 15, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.process_excluding_downtimes_slatimes(begin, end, sla_begin, sla_end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 1 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 4 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_two_days_excluding_slatimes_with_two_downtimes():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 23, 0)

        sla_begin = datetime.time(8)
        sla_end = datetime.time(19)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 15, 0)

        dt2_start = datetime.datetime(2015, 10, 25, 14, 0)
        dt2_end   = datetime.datetime(2015, 10, 25, 15, 0)


        downtimes = [(dt1_start, dt1_end), (dt2_start, dt2_end)]
        duration_and_days = list(duration.process_excluding_downtimes_slatimes(begin, end, sla_begin, sla_end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 5 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == (6 + 4) * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_two_days_excluding_slatimes_with_two_overspanning_downtimes():
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 23, 0)

        sla_begin = datetime.time(8)
        sla_end = datetime.time(19)

        dt1_start = datetime.datetime(2015, 10, 23, 10, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 15, 0)

        dt2_start = datetime.datetime(2015, 10, 25, 18, 0)
        dt2_end   = datetime.datetime(2015, 10, 26, 10, 0)


        downtimes = [(dt1_start, dt1_end), (dt2_start, dt2_end)]
        duration_and_days = list(duration.process_excluding_downtimes_slatimes(begin, end, sla_begin, sla_end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 4 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 10 * 3600
        assert date == datetime.date(2015, 10, 25)
