# ### RUN THIS in VirtualEnv with py.test *filename*

import datetime
import duration

import unittest as test

class DurationTest(test.TestCase):
    def test_single_day(self):
        begin = datetime.datetime(2015, 12, 6, 18, 0)
        end   = datetime.datetime(2015, 12, 6, 19, 0)

        duration_and_days = list(duration.duration.process(begin, end))
        assert len(duration_and_days) == 1

        dur, date = duration_and_days[0]

        assert dur == 3600
        assert date == datetime.date(2015, 12, 6)


    def test_two_days(self):
        begin = datetime.datetime(2015, 12, 6, 18, 0)
        end   = datetime.datetime(2015, 12, 7, 19, 0)

        duration_and_days = list(duration.duration.process(begin, end))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 12, 6)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 19 * 3600
        assert date == datetime.date(2015, 12, 7)


    def test_three_days(self):
        begin = datetime.datetime(2015, 12, 6, 18, 0)
        end   = datetime.datetime(2015, 12, 8, 20, 0)

        duration_and_days = list(duration.duration.process(begin, end))
        assert len(duration_and_days) == 3

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 12, 6)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 24 * 3600
        assert date == datetime.date(2015, 12, 7)

        # Day 2
        dur, date = duration_and_days[2]
        assert dur == 20 * 3600
        assert date == datetime.date(2015, 12, 8)


    def test_leap_hour(self):
        return  # test does not work as of yet
        begin = datetime.datetime(2015, 10, 24, 18, 0)
        end   = datetime.datetime(2015, 10, 25, 20, 0)

        duration_and_days = list(duration.duration.process(begin, end))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 19 * 3600  # 1 hour less than normal - clocks should go backwards 1h
        assert date == datetime.date(2015, 10, 25)


    def test_single_day_excluding_sla(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 24, 23, 0)
        sla_begin = datetime.time(8)
        sla_end   = datetime.time(19)

        duration_and_days = list(duration.duration.process_wo_slatimes(begin, end, sla_begin, sla_end))
        assert len(duration_and_days) == 1

        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_two_days_excluding_sla(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)
        sla_begin = datetime.time(8)
        sla_end   = datetime.time(19)

        duration_and_days = list(duration.duration.process_wo_slatimes(begin, end, sla_begin, sla_end))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 5 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_three_days_excluding_sla(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 26, 13, 0)
        sla_begin = datetime.time(8)
        sla_end   = datetime.time(19)

        duration_and_days = list(duration.duration.process_wo_slatimes(begin, end, sla_begin, sla_end))
        assert len(duration_and_days) == 3

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 11 * 3600
        assert date == datetime.date(2015, 10, 25)

        # Day 3
        dur, date = duration_and_days[2]
        assert dur == 5 * 3600
        assert date == datetime.date(2015, 10, 26)


    def test_excluding_one_downtime(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 24, 23, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 20, 0)

        downtimes = (dt1_start, dt1_end)
        duration_and_days = list(duration.duration.process_wo_downtimes(begin, end, [downtimes]))
        assert len(duration_and_days) == 1

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 4 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_excluding_two_downtimes(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 24, 23, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 15, 0)

        dt2_start = datetime.datetime(2015, 10, 24, 20, 0)
        dt2_end   = datetime.datetime(2015, 10, 24, 21, 0)

        downtimes = [(dt1_start, dt1_end), (dt2_start, dt2_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 1

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 8 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_two_days_excluding_one_downtime(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 15, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 10 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 13 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_two_days_excluding_multiple_day_downtime(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 25, 1, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 1 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 12 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_two_days_excluding_multiple_downtimes(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 25, 1, 0)

        dt2_start = datetime.datetime(2015, 10, 25, 12, 0)
        dt2_end   = datetime.datetime(2015, 10, 25, 13, 0)

        downtimes = [(dt1_start, dt1_end), (dt2_start, dt2_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 1 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 11 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_downtime_before_outtage(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 11, 0)
        dt1_end   = datetime.datetime(2015, 10, 25, 0, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 0 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 13 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_downtime_end_before_outtage(self):
        begin = datetime.datetime(2015, 10, 24, 12, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 11, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 12, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 12 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 13 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_downtime_longer_than_outtage(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 25, 12, 0)
        dt1_end   = datetime.datetime(2015, 10, 25, 15, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 11 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 12 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_downtime_overspanning_outtage(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 13, 0)

        dt1_start = datetime.datetime(2015, 10, 24, 12, 0)
        dt1_end   = datetime.datetime(2015, 10, 25, 15, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes(begin, end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 0 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 0 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_one_day_excluding_sla_downtime(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 24, 23, 0)

        sla_begin = datetime.time(8)
        sla_end = datetime.time(19)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 15, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes_slatimes(begin, end, sla_begin, sla_end, downtimes))
        assert len(duration_and_days) == 1

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 5 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_two_days_excluding_sla_downtime(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 23, 0)

        sla_begin = datetime.time(8)
        sla_end = datetime.time(19)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 15, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes_slatimes(begin, end, sla_begin, sla_end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 5 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 11 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_two_days_excluding_slatimes_with_long_downtime(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 23, 0)

        sla_begin = datetime.time(8)
        sla_end = datetime.time(19)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 25, 15, 0)

        downtimes = [(dt1_start, dt1_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes_slatimes(begin, end, sla_begin, sla_end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 1 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 4 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_two_days_excluding_slatimes_with_two_downtimes(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 23, 0)

        sla_begin = datetime.time(8)
        sla_end = datetime.time(19)

        dt1_start = datetime.datetime(2015, 10, 24, 14, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 15, 0)

        dt2_start = datetime.datetime(2015, 10, 25, 14, 0)
        dt2_end   = datetime.datetime(2015, 10, 25, 15, 0)


        downtimes = [(dt1_start, dt1_end), (dt2_start, dt2_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes_slatimes(begin, end, sla_begin, sla_end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 5 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == (6 + 4) * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_two_days_excluding_slatimes_with_two_overspanning_downtimes(self):
        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 25, 23, 0)

        sla_begin = datetime.time(8)
        sla_end = datetime.time(19)

        dt1_start = datetime.datetime(2015, 10, 23, 10, 0)
        dt1_end   = datetime.datetime(2015, 10, 24, 15, 0)

        dt2_start = datetime.datetime(2015, 10, 25, 18, 0)
        dt2_end   = datetime.datetime(2015, 10, 26, 10, 0)


        downtimes = [(dt1_start, dt1_end), (dt2_start, dt2_end)]
        duration_and_days = list(duration.duration.process_wo_downtimes_slatimes(begin, end, sla_begin, sla_end, downtimes))
        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 4 * 3600
        assert date == datetime.date(2015, 10, 24)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 10 * 3600
        assert date == datetime.date(2015, 10, 25)

    def test_non_sla_days(self):
        non_sla_days = (5, 6)
        begin = datetime.datetime(2016, 4, 10, 10, 0)
        end   = datetime.datetime(2016, 4, 10, 22, 0)

        sla_begin = datetime.time(8)
        sla_end   = datetime.time(19)

        duration_and_days = list(duration.duration.process_wo_slatimes(begin, end, sla_begin, sla_end, non_sla_days))
        assert len(duration_and_days) == 0

    def test_basic_process_within_multiple_slatimes(self):
        timeframes = {}
        timefrom = datetime.time(8)
        timeto   = datetime.time(19)

        timeframes[0] = [(timefrom, timeto)]
        timeframes[1] = [(timefrom, timeto)]
        timeframes[2] = [(timefrom, timeto)]
        timeframes[3] = [(timefrom, timeto)]
        timeframes[4] = [(timefrom, timeto)]
        timeframes[5] = [(timefrom, timeto)]
        timeframes[6] = [(timefrom, timeto)]

        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 24, 23, 0)

        duration_and_days = list(duration.duration.process_within_multiple_slatimes(
            begin,
            end,
            timeframes))

        assert len(duration_and_days) == 1

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 6 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_two_timeframes(self):
        timeframes = {}
        timefrom = datetime.time(8)
        timeto   = datetime.time(9)

        timefrom2 = datetime.time(10)
        timeto2   = datetime.time(11)

        timeframes[0] = [(timefrom, timeto), (timefrom2, timeto2)]
        timeframes[1] = [(timefrom, timeto), (timefrom2, timeto2)]
        timeframes[2] = [(timefrom, timeto), (timefrom2, timeto2)]
        timeframes[3] = [(timefrom, timeto), (timefrom2, timeto2)]
        timeframes[4] = [(timefrom, timeto), (timefrom2, timeto2)]
        timeframes[5] = [(timefrom, timeto), (timefrom2, timeto2)]
        timeframes[6] = [(timefrom, timeto), (timefrom2, timeto2)]

        begin = datetime.datetime(2015, 10, 24, 7, 0)
        end   = datetime.datetime(2015, 10, 24, 23, 0)

        duration_and_days = list(duration.duration.process_within_multiple_slatimes(
            begin,
            end,
            timeframes))

        assert len(duration_and_days) == 1

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 2 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_timeframe_before_outtage(self):
        timeframes = {}
        timefrom = datetime.time(8)
        timeto   = datetime.time(9)

        timeframes[0] = [(timefrom, timeto)]
        timeframes[1] = [(timefrom, timeto)]
        timeframes[2] = [(timefrom, timeto)]
        timeframes[3] = [(timefrom, timeto)]
        timeframes[4] = [(timefrom, timeto)]
        timeframes[5] = [(timefrom, timeto)]
        timeframes[6] = [(timefrom, timeto)]

        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 24, 23, 0)

        duration_and_days = list(duration.duration.process_within_multiple_slatimes(
            begin,
            end,
            timeframes))

        assert len(duration_and_days) == 1

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 0 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_timeframe_overlapping_outtage_start(self):
        timeframes = {}
        timefrom = datetime.time(8)
        timeto   = datetime.time(18)

        timeframes[0] = [(timefrom, timeto)]
        timeframes[1] = [(timefrom, timeto)]
        timeframes[2] = [(timefrom, timeto)]
        timeframes[3] = [(timefrom, timeto)]
        timeframes[4] = [(timefrom, timeto)]
        timeframes[5] = [(timefrom, timeto)]
        timeframes[6] = [(timefrom, timeto)]

        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 24, 23, 0)

        duration_and_days = list(duration.duration.process_within_multiple_slatimes(
            begin,
            end,
            timeframes))

        assert len(duration_and_days) == 1

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 5 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_timeframe_overlapping_outtage_end(self):
        timeframes = {}
        timefrom = datetime.time(14)
        timeto   = datetime.time(19)

        timeframes[0] = [(timefrom, timeto)]
        timeframes[1] = [(timefrom, timeto)]
        timeframes[2] = [(timefrom, timeto)]
        timeframes[3] = [(timefrom, timeto)]
        timeframes[4] = [(timefrom, timeto)]
        timeframes[5] = [(timefrom, timeto)]
        timeframes[6] = [(timefrom, timeto)]

        begin = datetime.datetime(2015, 10, 24, 13, 0)
        end   = datetime.datetime(2015, 10, 24, 18, 0)

        duration_and_days = list(duration.duration.process_within_multiple_slatimes(
            begin,
            end,
            timeframes))

        assert len(duration_and_days) == 1

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 4 * 3600
        assert date == datetime.date(2015, 10, 24)

    def test_timeframe_for_specific_days(self):
        timeframes = {}
        timefrom = datetime.time(0)
        timeto   = datetime.time(23)

        timefrom2 = datetime.time(9)
        timeto2   = datetime.time(18)

        timeframes[0] = [(timefrom, timeto)]
        timeframes[1] = [(timefrom, timeto)]
        timeframes[2] = [(timefrom, timeto)]
        timeframes[3] = [(timefrom, timeto)]
        timeframes[4] = [(timefrom, timeto)]
        timeframes[5] = [(timefrom2, timeto2)]
        timeframes[6] = [(timefrom2, timeto2)]

        begin = datetime.datetime(2016, 4, 10, 7, 0)
        end   = datetime.datetime(2016, 4, 10, 12, 0)

        duration_and_days = list(duration.duration.process_within_multiple_slatimes(
            begin,
            end,
            timeframes))

        assert len(duration_and_days) == 1

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 3 * 3600
        assert date == datetime.date(2016, 4, 10)

    def test_timeframe_over_two_separate_timeframes(self):
        timeframes = {}
        timefrom = datetime.time(0)
        timeto   = datetime.time(23)

        timefrom2 = datetime.time(9)
        timeto2   = datetime.time(18)

        timeframes[0] = [(timefrom, timeto)]
        timeframes[1] = [(timefrom, timeto)]
        timeframes[2] = [(timefrom, timeto)]
        timeframes[3] = [(timefrom, timeto)]
        timeframes[4] = [(timefrom, timeto)]
        timeframes[5] = [(timefrom2, timeto2)]
        timeframes[6] = [(timefrom2, timeto2)]

        begin = datetime.datetime(2016, 4, 8, 7, 0)
        end   = datetime.datetime(2016, 4, 9, 12, 0)

        duration_and_days = list(duration.duration.process_within_multiple_slatimes(
            begin,
            end,
            timeframes))

        assert len(duration_and_days) == 2

        # Day 1
        dur, date = duration_and_days[0]
        assert dur == 16 * 3600
        assert date == datetime.date(2016, 4, 8)

        # Day 2
        dur, date = duration_and_days[1]
        assert dur == 3 * 3600
        assert date == datetime.date(2016, 4, 9)


    def test_empty_timeframes(self):
        timeframes = {}

        begin = datetime.datetime(2016, 4, 8, 7, 0)
        end   = datetime.datetime(2016, 4, 8, 12, 0)

        duration_and_days = list(duration.duration.process_within_multiple_slatimes(
            begin,
            end,
            timeframes))

        assert len(duration_and_days) == 0
