"""Functions to run the rpc server"""
# pylint:disable=bad-whitespace,import-error

import os

if 'DJANGO_SETTINGS_MODULE' not in os.environ:
    os.environ['DJANGO_SETTINGS_MODULE'] = "symonitoring_rpcserver.settings"

# System imports
import django        # noqa
import sys           # noqa
import time          # noqa
import setproctitle  # noqa
import traceback     # noqa

# Project imports
from symonitoring_rpcserver.version     import get_version  # noqa
from django.conf                        import settings     # noqa
from django.apps                        import AppConfig    # noqa
from symonitoring_rpcserver.apiregistry import api_handler  # noqa
import syrpc.server                     as srv              # noqa
import symonitoring_rpcserver.common    as cmn              # noqa

django.setup()

# TODO: clean this up! we shouldn't import the api but use AppConfig
import symonitoring_rpcserver.api  # noqa
import symonitoring_bpe.api        # noqa
import symonitoring_reporting.api  # noqa


def server(sleep=None, timeout=10, runner=False):
    """symonitoring_rpc Server main routine"""
    rpc_server = srv.Server(settings.AMQ_SETTINGS)
    if runner:
        while True:
            try:
                run_server(rpc_server=rpc_server, timeout=timeout, sleep=sleep)
            except srv.EmptyException:
                cmn.lg.info("Timeout reached before server got a request")
            except srv.QueueNotFoundException:
                cmn.lg.exception("Queue does not exist")
    else:
        run_server(rpc_server=rpc_server, timeout=timeout, sleep=sleep)


def server_runner():
    """symonitoring_rpc Server runner routine"""
    server(runner=True, timeout=None)


def run_server(rpc_server, timeout, sleep=None):
    """Run the server

    :param rpc_server: The server instance to be used for sending the request
    :type  rpc_server: symonitoring_rpc.server.Server
    :param    timeout: A request timeout in seconds
    :type     timeout: int
    :param      sleep: Amount of seconds to wait before processing the response
    :type       sleep: float
    """
    cmn.lg.debug('Running server...')
    callable_name = None
    result_id     = None
    data          = None
    try:
        # Get the name of the callable and request id & data associated with it
        callable_name, result_id, data = rpc_server.get_request(
            timeout=timeout
        )

        # Log the request's id
        cmn.lg.info("Server received request: %s", result_id)

        # Go to sleep if needed
        if sleep is not None:
            cmn.lg.debug("Sleeping for %d seconds...", sleep)
            time.sleep(sleep)

        #  TODO: register get_version as @api_callable
        if callable_name == 'get_version':
            res = {'version': get_version()}
        else:
            # Get the callable from the api_registry and call it
            res = api_handler(callable_name)(**data)

        # Log the callable's name and the data associated with it
        cmn.lg.debug("Server processed type: {0} - {1}".format(
            callable_name,
            data
        ))

    except Exception as e:
        res = {
            'exception': {
                'code':      e.args[0],
                'type':      e.__class__.__name__,
                'message':   str(e),  # .args[1],
                'traceback': traceback.format_exc(),
            }
        }

        # Log the exception
        cmn.lg.exception(
            'An exception occured in %s: %s (%s)',
            callable_name,
            e,
            e.__class__.__name__
        )

    # put the result
    rpc_server.put_result(result_id=result_id, data=res)
    cmn.lg.debug("Server put result: {0} ({1})".format(res, result_id))


setproctitle.setproctitle("syrpcserver")
