# -*- coding: UTF-8 -*-
# pylint: disable=missing-docstring

"""Tests for symonitoring_bpe"""

# System imports
import mock
from django.test              import TestCase

# Project imports
import syrpc.server           as srv
import symonitoring_rpcserver


class ServerTests(TestCase):
    """Test symonitoring_server functions"""

    def test_server(self):
        patch_srv = mock.patch('syrpc.server.Server')
        patch_run_server = mock.patch('symonitoring_rpcserver.run_server')
        with patch_srv as mock_srv, patch_run_server as mock_run_server:
            symonitoring_rpcserver.server(sleep=None, timeout=5, runner=False)
            self.assertTrue(mock_srv.called)
            mock_run_server.assert_called_once_with(
                rpc_server=mock_srv(),
                timeout=5,
                sleep=None
            )

    def test_server_runner(self):
        with mock.patch('symonitoring_rpcserver.server') as mock_server:
            symonitoring_rpcserver.server_runner()
            mock_server.assert_called_once_with(runner=True, timeout=None)

    def test_run_server_get_application_list(self):
        rpc_server = mock.Mock()
        rpc_server.get_request.side_effect = ('get_application_list', 1, {},),
        get_application_list = mock.Mock()
        get_application_list.side_effect = lambda: {}
        patch_get_list = mock.patch(
            'symonitoring_bpe.api.API.get_application_list',
            get_application_list
        )
        with patch_get_list as mock_get_list:
            symonitoring_rpcserver.run_server(rpc_server, None, None)
            rpc_server.get_request.assert_called_once_with(timeout=None)
            self.assertTrue(rpc_server.put_result.called)
            self.assertTrue(mock_get_list.called)

    def test_run_server_get_application(self):
        rpc_server = mock.Mock()
        rpc_server.get_request.side_effect = (
            'get_application',
            1,
            {
                'application_id': 1
            },
        ),
        patch_get = mock.patch(
            'symonitoring_bpe.api.API.get_application',
            mock.Mock(side_effect=lambda x: {}),
        )
        with patch_get as mock_get:
            symonitoring_rpcserver.run_server(rpc_server, None, None)
            rpc_server.get_request.assert_called_once_with(timeout=None)
            self.assertTrue(rpc_server.put_result.called)
            self.assertTrue(mock_get.called)

    def test_run_server_get_application_fail(self):
        rpc_server = mock.Mock()
        rpc_server.get_request.side_effect = (
            'get_application',
            1,
            {
                'application_id': None
            },
        ),
        e = IndexError('Application with id {} does not exist'.format(
            None
        ))
        patch_get = mock.patch(
            'symonitoring_bpe.api.API.get_application',
            mock.Mock(side_effect=e),
        )
        patch_exp = mock.patch(
            'symonitoring_rpcserver.cmn.lg.exception'
        )
        with patch_get as mock_get, patch_exp as mock_exp:
            symonitoring_rpcserver.run_server(rpc_server, None, None)
            self.assertTrue(mock_exp.called)
            rpc_server.get_request.assert_called_once_with(timeout=None)
            self.assertTrue(rpc_server.put_result.called)
            self.assertTrue(mock_get.called)

    def test_run_server_empty_exception(self):
        rpc_server = mock.Mock()
        rpc_server.get_request.side_effect = srv.EmptyException()
        patch_exp = mock.patch(
            'symonitoring_rpcserver.cmn.lg.exception'
        )
        with patch_exp as mock_exp:
            symonitoring_rpcserver.run_server(rpc_server, 0, None)
            self.assertTrue(mock_exp.called)
        self.assertTrue(rpc_server.get_request.called)

    def test_run_server_queue_not_found_exception(self):
        rpc_server = mock.Mock()
        rpc_server.get_request.side_effect = srv.QueueNotFoundException()
        patch_exp = mock.patch(
            'symonitoring_rpcserver.cmn.lg.exception'
        )
        with patch_exp as mock_exp:
            symonitoring_rpcserver.run_server(rpc_server, 0, None)
            self.assertTrue(mock_exp.called)
        self.assertTrue(rpc_server.get_request.called)

    def test_run_server_sleep(self):
        rpc_server = mock.Mock()
        rpc_server.get_request.side_effect = ('get_application_list', 0, {},),
        with mock.patch('time.sleep') as sleep_patched:
            symonitoring_rpcserver.run_server(rpc_server, 0, 9001)
            self.assertTrue(sleep_patched.called)
            sleep_patched.assert_called_once_with(9001)

    def test_run_server_invalid_type(self):
        rpc_server = mock.Mock()
        rpc_server.get_request.side_effect = (None, 0, {},),
        symonitoring_rpcserver.run_server(rpc_server, 0, None)
        self.assertTrue(rpc_server.put_result.called)
        _, _, kwargs = rpc_server.method_calls[1]
        self.assertIn('data', kwargs)
        self.assertIn('exception', kwargs['data'])
        self.assertIn('type', kwargs['data']['exception'])
        self.assertEqual(
            'InvalidTypeException',
            kwargs['data']['exception']['type']
        )
